// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file StdMpi.h */

#ifndef AMDIS_STDMPI_H
#define AMDIS_STDMPI_H

#include <map>
#include "mpi.h"
#include "MeshStructure.h"

namespace AMDiS {

  int intSizeOf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data)
  {
    return data.size() * 2;
  }

  int intSizeOf(std::vector<MeshStructure> &data)
  {
    int s = 0;
    for (int i = 0; i < data.size(); i++)
      s += data[i].getCode().size() + 2;

    return s;
  }

  void makeBuf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data, int* buf)
  {
    int i = 0;
    for (std::map<const DegreeOfFreedom*, DegreeOfFreedom>::iterator it = data.begin();
	 it != data.end(); ++it) {
      buf[i++] = *(it->first);
      buf[i++] = it->second;
    }
  }
  
  void makeFromBuf(std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &data, 
		   int *buf, int bufSize)
  {    
    if (bufSize == 0)
      return;

    data.clear();
    data.reserve(bufSize / 2);

    int i = 0;
    do {
      data.push_back(std::make_pair(buf[i], buf[i + 1]));
      i += 2;
    } while (i < bufSize);
  }

  void makeBuf(std::vector<MeshStructure> &data, unsigned long int *buf)
  {
    int pos = 0;
    for (int i = 0; i < data.size(); i++) {
      buf[pos++] = data[i].getCode().size();
      buf[pos++] = data[i].getNumElements();
      for (int j = 0; j < data[i].getCode().size(); j++)
	buf[pos++] = data[i].getCode()[j];
    }
  }

  void makeFromBuf(std::vector<MeshStructure> &data, unsigned long int *buf, int bufSize)
  {
    int pos = 0;
    
    while (pos < bufSize) {       
      int codeSize = buf[pos++];
      int nElements = buf[pos++];
      std::vector<unsigned long int> code;
      code.resize(codeSize);
      for (int i = 0; i < codeSize; i++)
	code[i] = buf[pos++];

      MeshStructure meshCode;
      meshCode.init(code, nElements);

      data.push_back(meshCode);
    }	
  }

  template<typename SendT, typename RecvT>
  class StdMpi 
  {
  public:
    StdMpi(MPI::Intracomm &comm) :
      mpiComm(comm),
      commPrepared(false),
      exchangeDataSize(true)
    {}

    void prepareCommunication(bool b)
    {
      sendData.clear();
      recvData.clear();
            
      exchangeDataSize = b;
      commPrepared = true;
    }

    void send(int toRank, SendT& data)
    {
      FUNCNAME("StdMpi::send()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      sendData[toRank] = data;
      sendDataSize[toRank] = intSizeOf(data);
    }

    void send(std::map<int, SendT>& data)
    {
      FUNCNAME("StdMpi::send()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      for (typename std::map<int, SendT>::iterator it = data.begin(); 
	   it != data.end(); ++it) {
	sendData[it->first] = it->second;
	sendDataSize[it->first] = intSizeOf(it->second);
      }
    }

    void recv(int fromRank, int size = -1)
    {
      FUNCNAME("StdMpi::recv()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      recvDataSize[fromRank] = size;
    }

    std::map<int, RecvT>& getRecvData()
    {
      return recvData;
    }

    void commDataSize()
    {
      MPI::Request request[sendData.size() + recvDataSize.size()];
      std::vector<int> sendBuffers, recvBuffers;
      int requestCounter = 0;

      int i = 0;
      for (typename std::map<int, int>::iterator sendIt = sendDataSize.begin();
	   sendIt != sendDataSize.end(); ++sendIt) {
	sendBuffers.push_back(sendIt->second);
	request[requestCounter++] =
	  mpiComm.Isend(&(sendBuffers[i]), 1, MPI_INT, sendIt->first, 0);
	i++;
      }

      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt)
	request[requestCounter++] =
	  mpiComm.Irecv(&(recvIt->second), 1, MPI_INT, recvIt->first, 0);
      
      MPI::Request::Waitall(requestCounter, request);      
    }

    template<class T>
    void startCommunication()
    {
      FUNCNAME("StdMpi::startCommunication()");

      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      if (exchangeDataSize)
	commDataSize();

      MPI::Request request[sendData.size() + recvDataSize.size()];
      int requestCounter = 0;
      std::vector<T*> sendBuffers, recvBuffers;

      for (typename std::map<int, SendT>::iterator sendIt = sendData.begin();
	   sendIt != sendData.end(); ++sendIt) {
	int bufferSize = sendDataSize[sendIt->first];
	T* buf = new T[bufferSize];
	makeBuf(sendIt->second, buf);

  	request[requestCounter++] = 
  	  mpiComm.Isend(buf, bufferSize, MPI_INT, sendIt->first, 0);

	sendBuffers.push_back(buf);
      }

      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt) {
	T* buf = new T[recvIt->second];

  	request[requestCounter++] =
  	  mpiComm.Irecv(buf, recvIt->second, MPI_INT, recvIt->first, 0);

	recvBuffers.push_back(buf);	
      }

      MPI::Request::Waitall(requestCounter, request);

      for (int i = 0; i < static_cast<int>(sendBuffers.size()); i++)
	delete [] sendBuffers[i];
      sendBuffers.clear();

      int i = 0;
      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt) {
	makeFromBuf(recvData[recvIt->first], recvBuffers[i], recvIt->second);
	delete [] recvBuffers[i];
	i++;
      }      
    }

  protected:
    ///
    MPI::Intracomm mpiComm;

    ///
    std::map<int, SendT> sendData;

    ///
    std::map<int, RecvT> recvData;

    std::map<int, int> sendDataSize;

    std::map<int, int> recvDataSize;

    bool commPrepared;

    bool exchangeDataSize;
  };

}

#endif
